# =================================================================
#
# Authors: Ricardo Garcia Silva <ricardo.garcia.silva@gmail.com>
#
# Copyright (c) 2016 Ricardo Garcia Silva
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================
"""pytest configuration file for functional tests"""

import codecs
from collections import namedtuple
import logging
import os
import re
import configparser

import apipkg
import pytest

from pycsw.core import admin
from pycsw.core.config import StaticContext

apipkg.initpkg("optionaldependencies", {
    "psycopg2": "psycopg2",
})

from optionaldependencies import psycopg2  # NOQA: E402

TESTS_ROOT = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

SuiteDirs = namedtuple("SuiteDirs", [
    "get_tests_dir",
    "post_tests_dir",
    "data_tests_dir",
    "expected_results_dir",
    "export_tests_dir",
])


def pytest_generate_tests(metafunc):
    """Parametrize tests programmatically.

    This function scans the filesystem directories under
    ``tests/functionaltests/suites`` and automatically generates pytest tests
    based on the available test suites. Each suite directory has the
    following structure:

    * A mandatory ``default.cfg`` file specifying the configuration for the
      pycsw instance to use in the tests of the suite.

    * An optional ``get/`` subdirectory containing a ``requests.txt`` file
      with any HTTP GET requests for which to generate tests for. Each request
      is specified in a new line, with the following pattern:

      * <test_name>,<request_query_string>

    * An optional ``post/`` subdirectory containing files that are used as the
      payload for HTTP POST requests. The name of each file is used as the name
      of the test (without the file's extension);

    * An optional ``data/`` subdirectory. This directory, if present, indicates
      that the suite uses a custom database. The database is populated with any
      additional files that are contained inside this directory. If the
      ``data`` directory does not exist then the suite's tests will use the
      CITE database;

    * An ``expected/`` subdirectory containing a file for each of the expected
      test outcomes.

    The tests are autogenerated by parametrizing the
    ``tests/functionaltests/test_suites_functional::test_suites`` function

    Notes
    -----

    Check pytest's documentation for information on autogenerating
    parametrized tests for further details on how the
    ``pytest_generate_tests`` function can be used:

    http://pytest.org/latest/parametrize.html#basic-pytest-generate-tests-example

    """

    global TESTS_ROOT
    if metafunc.function.__name__ == "test_suites":
        suites_root_dir = os.path.join(TESTS_ROOT, "functionaltests", "suites")
        suite_names = os.listdir(suites_root_dir)
        arg_values = []
        test_ids = []
        logging.basicConfig(level=getattr(
            logging, metafunc.config.getoption("--pycsw-loglevel").upper()))
        if metafunc.config.getoption("--database-backend") == "postgresql":
            _recreate_postgresql_database(metafunc.config)
        for suite in suite_names:
            suite_dir = os.path.join(suites_root_dir, suite)
            config_path = os.path.join(suite_dir, "default.cfg")
            if not os.path.isfile(config_path):
                print("Directory {0!r} does not have a suite "
                      "configuration file".format(suite_dir))
                continue
            print("Generating tests for suite {0!r}...".format(suite))
            normalize_ids = True if suite in ("harvesting",
                                              "manager") else False
            suite_dirs = _get_suite_dirs(suite)
            if suite_dirs.post_tests_dir is not None:
                post_argvalues, post_ids = _get_post_parameters(
                    post_tests_dir=suite_dirs.post_tests_dir,
                    expected_tests_dir=suite_dirs.expected_results_dir,
                    config_path=config_path,
                    suite_name=suite,
                    normalize_ids=normalize_ids,
                )
                arg_values.extend(post_argvalues)
                test_ids.extend(post_ids)
            if suite_dirs.get_tests_dir is not None:
                get_argvalues, get_ids = _get_get_parameters(
                    get_tests_dir=suite_dirs.get_tests_dir,
                    expected_tests_dir=suite_dirs.expected_results_dir,
                    config_path=config_path,
                    suite_name=suite,
                    normalize_ids=normalize_ids,
                )
                arg_values.extend(get_argvalues)
                test_ids.extend(get_ids)

        metafunc.parametrize(
            argnames=["configuration", "request_method", "request_data",
                      "expected_result", "normalize_identifier_fields",],
            argvalues=arg_values,
            indirect=["configuration"],
            ids=test_ids,
        )


@pytest.fixture()
def test_identifier(request):
    """Extract a meaningful identifier from the request's node."""
    return re.search(r"[\w_]+\[(.*)\]", request.node.name).group(1)


@pytest.fixture()
def use_xml_canonicalisation(request):
    return not request.config.getoption("--functional-prefer-diffs")


@pytest.fixture()
def save_results_directory(request):
    return request.config.getoption("--functional-save-results-directory")


@pytest.fixture()
def configuration(request, tests_directory, log_level):
    """Configure a suite for execution in tests.

    This function is executed once for each individual test request, after
    tests have been collected.

    The configuration file for each test suite is read into memory. Some
    configuration parameters, like the repository's url and table name are
    adjusted. The suite's repository is also created, if needed.

    Parameters
    ----------
    request: pytest.fixtures.FixtureRequest
    tests_directory: py.path.local
        Directory created by pytest where any test artifacts are to be saved
    log_level: str
        Log level for the pycsw server instance that will be created during
        tests.

    """

    config_path = request.param
    config = configparser.ConfigParser()
    with codecs.open(config_path, encoding="utf-8") as fh:
        config.read_file(fh)
    suite_name = config_path.split(os.path.sep)[-2]
    suite_dirs = _get_suite_dirs(suite_name)
    data_dir = suite_dirs.data_tests_dir
    export_dir = suite_dirs.export_tests_dir
    if data_dir is not None:  # suite has its own database
        repository_url = _get_repository_url(request.config, suite_name,
                                             tests_directory)
    elif suite_name == 'opensearcheo':
        repository_url = _get_repository_url(request.config, "apiso",
                                             tests_directory)
    else:  # suite uses the CITE database
        data_dir, export_dir = _get_cite_suite_dirs()
        repository_url = _get_repository_url(request.config, "cite",
                                             tests_directory)
    table_name = _get_table_name(suite_name, config, repository_url)
    if not _repository_exists(repository_url, table_name):
        _initialize_database(repository_url=repository_url,
                             table_name=table_name,
                             data_dir=data_dir,
                             test_dir=tests_directory,
                             export_dir=export_dir)
    config.set("server", "loglevel", log_level)
    config.set("server", "logfile", "")
    config.set("repository", "database", repository_url)
    config.set("repository", "table", table_name)
    return config


@pytest.fixture(scope="session", name="tests_directory")
def fixture_tests_directory(tmpdir_factory):
    """Create a temporary directory for each test session.

    This directory is typically situated under ``/tmp`` and is used to create
    eventual sqlite databases for each suite.

    This functionality is mostly provided by pytest's built-in
    ``tmpdir_factory`` fixture. More information on this is available at:

    http://doc.pytest.org/en/2.9.0/tmpdir.html#the-tmpdir-factory-fixture

    """

    tests_dir = tmpdir_factory.mktemp("functional_tests")
    return tests_dir


def _get_cite_suite_dirs():
    """Return the path to the data directory of the CITE test suite."""
    global TESTS_ROOT
    suites_root_dir = os.path.join(TESTS_ROOT, "functionaltests", "suites")
    suite_dir = os.path.join(suites_root_dir, "cite")
    data_tests_dir = os.path.join(suite_dir, "data")
    export_tests_dir = os.path.join(suite_dir, "export")
    data_dir = data_tests_dir if os.path.isdir(data_tests_dir) else None
    export_dir = export_tests_dir if os.path.isdir(export_tests_dir) else None
    return data_dir, export_dir


def _get_get_parameters(get_tests_dir, expected_tests_dir, config_path,
                        suite_name, normalize_ids):
    """Return the parameters suitable for parametrizing HTTP GET tests."""
    method = "GET"
    test_argvalues = []
    test_ids = []
    requests_file_path = os.path.join(get_tests_dir, "requests.txt")
    with open(requests_file_path) as fh:
        for line in fh:
            test_name, test_params = [i.strip() for i in
                                      line.partition(",")[::2]]
            expected_result_path = os.path.join(
                expected_tests_dir,
                "{method}_{name}.xml".format(method=method.lower(),
                                             name=test_name)
            )
            test_argvalues.append(
                (config_path, method, test_params, expected_result_path,
                 normalize_ids)
            )
            test_ids.append(
                "{suite}_{http_method}_{name}".format(
                    suite=suite_name, http_method=method.lower(),
                    name=test_name)
            )
    return test_argvalues, test_ids


def _get_post_parameters(post_tests_dir, expected_tests_dir, config_path,
                         suite_name, normalize_ids):
    """Return the parameters suitable for parametrizing HTTP POST tests."""
    method = "POST"
    test_argvalues = []
    test_ids = []
    # we are sorting the directory contents because the
    # `harvesting` suite requires tests to be executed in alphabetical order
    directory_contents = sorted(os.listdir(post_tests_dir))
    for request_file_name in directory_contents:
        request_path = os.path.join(post_tests_dir,
                                    request_file_name)
        expected_result_path = os.path.join(
            expected_tests_dir,
            "{method}_{filename}".format(
                method=method.lower(),
                filename=request_file_name
            )
        )
        test_argvalues.append(
            (config_path, method, request_path,
             expected_result_path, normalize_ids)
        )
        test_ids.append(
            "{suite}_{http_method}_{file_name}".format(
                suite=suite_name,
                http_method=method.lower(),
                file_name=os.path.splitext(
                    request_file_name)[0])
        )
    return test_argvalues, test_ids


def _get_repository_url(conf, suite_name, test_dir):
    """Return the repository_url for the input parameters.

    Returns
    -------
    repository_url: str
        SQLAlchemy URL for the repository in use.

    """

    db_type = conf.getoption("--database-backend")
    if db_type == "sqlite":
        repository_url = "sqlite:///{test_dir}/{suite}.db".format(
            test_dir=test_dir, suite=suite_name)
    elif db_type == "postgresql":
        repository_url = (
            "postgresql://{user}:{password}@{host}:{port}/{database}".format(
                user=conf.getoption("--database-user-postgresql"),
                password=conf.getoption("--database-password-postgresql"),
                host=conf.getoption("--database-host-postgresql"),
                port=conf.getoption("--database-port-postgresql"),
                database=conf.getoption("--database-name-postgresql"))
        )
    else:
        raise NotImplementedError
    return repository_url


def _get_suite_dirs(suite_name):
    """Get the paths to relevant suite directories.

    Parameters
    ----------
    suite_name: str
        Name of the site

    Returns
    -------
    SuiteDirs
        A four element named tuple with the input suite's relevant test
        directories.

    """

    global TESTS_ROOT
    suites_root_dir = os.path.join(TESTS_ROOT, "functionaltests", "suites")
    suite_dir = os.path.join(suites_root_dir, suite_name)
    data_tests_dir = os.path.join(suite_dir, "data")
    post_tests_dir = os.path.join(suite_dir, "post")
    get_tests_dir = os.path.join(suite_dir, "get")
    export_tests_dir = os.path.join(suite_dir, "export")
    expected_results_dir = os.path.join(suite_dir, "expected")
    data_dir = data_tests_dir if os.path.isdir(data_tests_dir) else None
    posts_dir = post_tests_dir if os.path.isdir(post_tests_dir) else None
    gets_dir = get_tests_dir if os.path.isdir(get_tests_dir) else None
    expected_dir = (expected_results_dir if os.path.isdir(
        expected_results_dir) else None)
    export_dir = export_tests_dir if os.path.isdir(export_tests_dir) else None
    return SuiteDirs(get_tests_dir=gets_dir,
                     post_tests_dir=posts_dir,
                     data_tests_dir=data_dir,
                     expected_results_dir=expected_dir,
                     export_tests_dir=export_tests_dir)


def _get_table_name(suite, config, repository_url):
    """Get the name of the table used to store records in the database.

    Parameters
    ----------
    suite: str
        Name of the suite.
    config: ConfigParser
        Configuration for the suite.
    repository_url: str
        SQLAlchemy URL for the repository in use.

    Returns
    -------
    str
        Name of the table to use in the database

    """

    if repository_url.startswith("sqlite"):
        result = config.get("repository", "table")
    elif repository_url.startswith("postgresql"):
        result = "{suite}_records".format(suite=suite)
    else:
        raise NotImplementedError
    return result


def _initialize_database(repository_url, table_name, data_dir, test_dir, export_dir):
    """Initialize database for tests.

    This function will create the database and load any test data that
    the suite may require.

    Parameters
    ----------
    repository_url: str
        URL for the repository, as used by SQLAlchemy engines
    table_name: str
        Name of the table that is to be used to store pycsw records
    data_dir: str
        Path to a directory that contains sample data records to be loaded
        into the database
    test_dir: str
        Directory where the database is to be created, in case of sqlite.
    export_dir: str
        Diretory where the exported records are to be saved, if any

    """

    print("Setting up {0!r} repository...".format(repository_url))
    if repository_url.startswith("postgresql"):
        extra_kwargs = {
            "create_sfsql_tables": True,
            "create_plpythonu_functions": False
        }
    else:
        extra_kwargs = {}
    admin.setup_db(database=repository_url, table=table_name, home=test_dir,
                   **extra_kwargs)
    if len(os.listdir(data_dir)) > 0:
        print("Loading database data...")
        loaded = admin.load_records(
            context=StaticContext(),
            database=repository_url,
            table=table_name,
            xml_dirpath=data_dir,
            recursive=True
        )
        admin.optimize_db(context=StaticContext(), database=repository_url, table=table_name)
        if export_dir is not None:
            # Attempt to export files
            exported = admin.export_records(
                context=StaticContext(),
                database=repository_url,
                table=table_name,
                xml_dirpath=export_dir
            )
            if len(loaded) != len(exported):
                raise ValueError(
                    "Loaded records (%s) is different from exported records (%s)" %
                    (len(loaded), len(exported))
                )
            # Remove the files that were exported since this was just a test
            for toremove in exported:
                os.remove(toremove)


def _parse_postgresql_repository_url(repository_url):
    """Parse a SQLAlchemy engine URL describing a postgresql database.

    Parameters
    ----------
    repository_url: str
        SQLAlchemy URL for the repository in use.

    Returns
    -------
    dict
        A mapping with the database's connection parameters.

    """

    info_re = re.search(r"postgresql://(?P<user>[\w_]+):(?P<password>.*?)@"
                        r"(?P<host>[\w_.]+):(?P<port>\d+)/"
                        r"(?P<database>[\w_]+)",
                        repository_url,
                        flags=re.UNICODE)
    try:
        db_info = info_re.groupdict()
    except AttributeError:
        raise RuntimeError("Could not parse repository url {0!r}".format(
            repository_url))
    else:
        return db_info


def _recreate_postgresql_database(configuration):
    """Recreate a postgresql database.

    This function will try to create a new postgresql database for testing
    purposes. If the database already exists it is deleted and then recreated.

    Parameters
    ----------
    configuration: _pytest.config.Config
        The configuration object used by pytest

    Raises
    ------
    RuntimeError
        If a connection to the postgresql server cannot be made

    """

    connection = psycopg2.connect(
        database="postgres",
        user=configuration.getoption("--database-user-postgresql"),
        password=configuration.getoption("--database-password-postgresql"),
        host=configuration.getoption("--database-host-postgresql"),
        port=configuration.getoption("--database-port-postgresql")
    )
    connection.set_isolation_level(
        psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
    cursor = connection.cursor()
    db_name = configuration.getoption("--database-name-postgresql")
    cursor.execute("DROP DATABASE IF EXISTS {database}".format(
        database=db_name))
    cursor.execute("CREATE DATABASE {database}".format(database=db_name))
    cursor.execute(
        "SELECT COUNT(1) FROM pg_available_extensions WHERE name='postgis'")
    postgis_available = bool(cursor.fetchone()[0])
    cursor.close()
    connection.close()
    if postgis_available:
        _create_postgresql_extension(configuration, extension="postgis")
    else:
        _create_postgresql_extension(configuration, extension="plpythonu")


def _create_postgresql_extension(configuration, extension):
    """Create a postgresql extension in a previously created database.

    Parameters
    ----------
    configuration: _pytest.config.Config
        The configuration object used by pytest
    extension: str
        Name of the extension to be created

    """

    connection = psycopg2.connect(
        database=configuration.getoption("--database-name-postgresql"),
        user=configuration.getoption("--database-user-postgresql"),
        password=configuration.getoption("--database-password-postgresql"),
        host=configuration.getoption("--database-host-postgresql"),
        port=configuration.getoption("--database-port-postgresql")
    )
    connection.set_isolation_level(
        psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
    cursor = connection.cursor()
    cursor.execute("CREATE EXTENSION {0}".format(extension))
    cursor.close()
    connection.close()


def _repository_exists(repository_url, table_name):
    """Test if the database already exists.

    Parameters
    ----------
    repository_url: str
        URL for the repository, as used by SQLAlchemy engines
    table_name: str
        Name of the table that is to be used to store pycsw records

    Returns
    -------
    bool
        Whether the repository exists or not.

    """

    if repository_url.startswith("sqlite"):
        repository_path = repository_url.replace("sqlite:///", "")
        result = os.path.isfile(repository_path)
    elif repository_url.startswith("postgresql"):
        db_info = _parse_postgresql_repository_url(repository_url)
        try:
            connection = psycopg2.connect(user=db_info["user"],
                                          password=db_info["password"],
                                          host=db_info["host"],
                                          port=db_info["port"],
                                          database=db_info["database"])
            cursor = connection.cursor()
            cursor.execute("SELECT COUNT(1) FROM {table_name}".format(
                table_name=table_name))
        except (psycopg2.OperationalError, psycopg2.ProgrammingError):
            # database or table does not exist yet
            result = False
        else:
            result = True
    else:
        raise NotImplementedError
    return result



# -*- coding: utf-8 -*-
# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#          Ricardo Garcia Silva <ricardo.garcia.silva@gmail.com>
#
# Copyright (c) 2015 Tom Kralidis
# Copyright (c) 2016 Ricardo Garcia Silva
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import os

from paver.easy import task, cmdopts, needs, \
    pushd, sh, call_task, path, info

DOCS = 'docs'
STAGE_DIR = '/tmp'


@task
def build_release():
    """Create release package"""
    pass


@task
def refresh_docs():
    """Build sphinx docs from scratch"""
    with pushd(DOCS):
        sh('make clean')
        sh('make html')


@task
@cmdopts([
    ('user=', 'u', 'OSGeo userid'),
])
def publish_docs(options):
    """Publish dev docs to production"""
    local_path = '_build/html'
    remote_host = 'pycsw.org'
    remote_path = '/osgeo/pycsw/pycsw-web/docs/latest'

    user = options.get('user', False)
    if not user:
        raise Exception('OSGeo userid required')

    call_task('refresh_docs')

    with pushd(DOCS):
        # change privs to be group writeable
        for root, dirs, files in os.walk(local_path):
            for dfile in files:
                os.chmod(os.path.join(root, dfile), 0o664)
            for ddir in dirs:
                os.chmod(os.path.join(root, ddir), 0o775)

        # copy documentation
        sh('scp -r %s%s* %s@%s:%s' % (local_path, os.sep, user, remote_host,
                                      remote_path))


@task
def gen_tests_html():
    """Generate tests/index.html for online testing"""
    with pushd('tests'):
        # ensure manager testsuite is writeable
        os.chmod(os.path.join('functionaltests', 'suites', 'manager', 'data'), 0o777)
        sh('python gen_html.py > index.html')


@task
@needs(['distutils.command.sdist'])
def publish_pypi():
    """Publish to PyPI"""
    pass


@task
def package():
    """Package a distribution .tar.gz/.zip"""

    import pycsw

    version = pycsw.__version__

    package_name = 'pycsw-%s' % version

    call_task('package_tar_gz', options={'package_name': package_name})


@task
@cmdopts([
    ('package_name=', 'p', 'Name of package'),
])
def package_tar_gz(options):
    """Package a .tar.gz distribution"""

    import tarfile

    package_name = options.get('package_name', None)

    if package_name is None:
        raise Exception('Package name required')

    filename = path('%s/%s.tar.gz' % (STAGE_DIR, package_name))

    if filename.exists():
        info('Package %s already exists' % filename)
        return

    with pushd(STAGE_DIR):
        stage_path = '%s/%s' % (STAGE_DIR, package_name)

        if not path(stage_path).exists():
            raise Exception('Directory %s does not exist' % stage_path)

        tar = tarfile.open(filename, 'w:gz')
        tar.add(package_name)
        tar.close()



# =================================================================
#
# Authors: Ricardo Garcia Silva <ricardo.garcia.silva@gmail.com>
#
# Copyright (c) 2017 Ricardo Garcia Silva
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================
"""Unit tests for pycsw.core.repository"""

import pytest

from pycsw.core import repository

pytestmark = pytest.mark.unit


@pytest.mark.parametrize("data, input_, predicate, distance, expected", [
    ("LINESTRING(0 0, 1 1)", "POINT(0.5 0.5)", "bbox", 0, "true"),
    ("LINESTRING(0 0, 1 1)", "POINT(2 2)", "bbox", 0, "false"),
    ("LINESTRING(0 0, 1 1)", "POINT(2 2)", "beyond", 1, "true"),
    ("LINESTRING(0 0, 1 1)", "POINT(2 2)", "beyond", 2, "false"),
    ("LINESTRING(0 0, 1 1)", "POINT(0.5 0.5)", "beyond", "false", "false"),
    ("LINESTRING(0 0, 1 1)", "POINT(0.5 0.5)", "contains", 0, "true"),
    ("LINESTRING(0 0, 1 1)", "POINT(2 2)", "contains", 0, "false"),
    ("LINESTRING(0 0, 1 1)", "LINESTRING(1 0, 0 1)", "crosses", 0, "true"),
    ("LINESTRING(0 0, 1 1)", "POINT(0.5 0.5)", "crosses", 0, "false"),
    ("POINT(1 1)", "POINT(1 1)", "equals", 0, "true"),
    ("LINESTRING(0 0, 1 1)", "POINT(0.5 0.5)", "equals", 0, "false"),
    ("LINESTRING(0 0, 1 1)", "POINT(0 0)", "touches", 0, "true"),
    ("LINESTRING(0 0, 1 1)", "POINT(0.5 0.5)", "touches", 0, "false"),
    ("POINT(0.5 0.5)", "LINESTRING(0 0, 1 1)", "within", 0, "true"),
    ("LINESTRING(0 0, 1 1)", "POINT(0.5 0.5)", "within", 0, "false"),
    (None, "POINT(0.5 0.5)", "within", 0, "false"),
    ("POINT(0.5 0.5)", None, "within", 0, "false"),
    (None, None, "within", 0, "false"),
    ("LINESTRING(0 0, 1 1)", "POINT(0.5 0.5)", "dwithin", "false", "false"),
])
def test_query_spatial(data, input_, predicate, distance, expected):
    result = repository.query_spatial(
        bbox_data_wkt=data,
        bbox_input_wkt=input_,
        predicate=predicate,
        distance=distance
    )
    assert result == expected



# -*- coding: utf-8 -*-
# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#          Angelos Tzotsos <tzotsos@gmail.com>
#          Ricardo Garcia Silva <ricardo.garcia.silva@gmail.com>
#
# Copyright (c) 2015 Tom Kralidis
# Copyright (c) 2015 Angelos Tzotsos
# Copyright (c) 2017 Ricardo Garcia Silva
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import os
import re
import datetime
import logging
import time

from urllib.request import Request, urlopen
from urllib.parse import urlparse
from shapely.wkt import loads
from owslib.util import http_post

from pycsw.core.etree import etree, PARSER

LOGGER = logging.getLogger(__name__)

# Global variables for spatial ranking algorithm
ranking_enabled = False
ranking_pass = False
ranking_query_geometry = ''

# Lookups for the secure_filename function
# https://github.com/pallets/werkzeug/blob/778f482d1ac0c9e8e98f774d2595e9074e6984d7/werkzeug/utils.py#L30-L31
_filename_ascii_strip_re = re.compile(r'[^A-Za-z0-9_.-]')
_windows_device_files = ('CON', 'AUX', 'COM1', 'COM2', 'COM3', 'COM4', 'LPT1',
                         'LPT2', 'LPT3', 'PRN', 'NUL')


def get_today_and_now():
    """Get the date, right now, in ISO8601"""
    return time.strftime('%Y-%m-%dT%H:%M:%SZ', time.localtime())


def datetime2iso8601(value):
    """Return a datetime value as ISO8601

    Parameters
    ----------
    value: datetime.date or datetime.datetime
        The temporal value to be converted

    Returns
    -------
    str
        A string with the temporal value in ISO8601 format.

    """

    if isinstance(value, datetime.datetime):
        if value == value.replace(hour=0, minute=0, second=0, microsecond=0):
            result = value.strftime("%Y-%m-%d")
        else:
            result = value.strftime("%Y-%m-%dT%H:%M:%SZ")
    else:  # value is a datetime.date
        result = value.strftime('%Y-%m-%d')
    return result


def get_time_iso2unix(isotime):
    """Convert ISO8601 to UNIX timestamp"""
    return int(time.mktime(time.strptime(
        isotime, '%Y-%m-%dT%H:%M:%SZ'))) - time.timezone


def get_version_integer(version):
    """Get an integer of the OGC version value x.y.z

    In case of an invalid version string this returns -1.

    Parameters
    ----------
    version: str
        The version string that is to be transformed into an integer

    Returns
    -------
    int
        The transformed version

    Raises
    ------
    RuntimeError
        When the input version is neither a string or None

    """

    try:
        xyz = version.split('.')
        if len(xyz) == 3:
            result = int(xyz[0]) * 10000 + int(xyz[1]) * 100 + int(xyz[2])
        else:
            result = -1
    except AttributeError as err:
        raise RuntimeError('%s' % str(err))
    return result


def nspath_eval(xpath, nsmap):
    """Return an etree friendly xpath.

    This function converts XPath expressions that use prefixes into
    their full namespace. This is the form expected by lxml [1]_.

    Parameters
    ----------
    xpath: str
        The XPath expression to be converted
    nsmap: dict

    Returns
    -------
    str
        The XPath expression using namespaces instead of prefixes.

    References
    ----------
    .. [1] http://lxml.de/tutorial.html#namespaces

    """

    out = []
    for node in xpath.split('/'):
        chunks = node.split(":")
        if len(chunks) == 2:
            prefix, element = node.split(':')
            out.append('{%s}%s' % (nsmap[prefix], element))
        elif len(chunks) == 1:
            out.append(node)
        else:
            raise RuntimeError("Invalid XPath expression: {0}".format(xpath))
    return '/'.join(out)


def wktenvelope2bbox(envelope):
    """returns bbox string of WKT ENVELOPE definition"""

    tmparr = [x.strip() for x in envelope.split('(')[1].split(')')[0].split(',')]
    bbox = '%s,%s,%s,%s' % (tmparr[0], tmparr[3], tmparr[1], tmparr[2])
    return bbox


def wkt2geom(ewkt, bounds=True):
    """Return Shapely geometry object based on WKT/EWKT

    Parameters
    ----------
    ewkt: str
        The geometry to convert, in Extended Well-Known Text format. More info
        on this format at [1]_
    bounds: bool
        Whether to return only the bounding box of the geometry as a tuple or
        the full shapely geometry instance

    Returns
    -------
    shapely.geometry.base.BaseGeometry or tuple
        Depending on the value of the ``bounds`` parameter, returns either 
        the shapely geometry instance or a tuple with the bounding box.

    References
    ----------
    .. [1] http://postgis.net/docs/ST_GeomFromEWKT.html

    """

    wkt = ewkt.split(";")[-1] if ewkt.find("SRID") != -1 else ewkt
    if wkt.startswith('ENVELOPE'):
        wkt = bbox2wktpolygon(wktenvelope2bbox(wkt))
    geometry = loads(wkt)
    return geometry.envelope.bounds if bounds else geometry


def bbox2wktpolygon(bbox):
    """Return OGC WKT Polygon of a simple bbox string

    Parameters
    ----------
    bbox: str
        The bounding box to convert to WKT.

    Returns
    -------
    str
        The bounding box's Well-Known Text representation.

    """

    if bbox.startswith('ENVELOPE'):
        bbox = wktenvelope2bbox(bbox)
    minx, miny, maxx, maxy = [float(coord) for coord in bbox.split(",")]
    return 'POLYGON((%.2f %.2f, %.2f %.2f, %.2f %.2f, %.2f %.2f, %.2f %.2f))' \
        % (minx, miny, minx, maxy, maxx, maxy, maxx, miny, minx, miny)


def transform_mappings(queryables, typename):
    """Transform metadata model mappings

    Parameters
    ----------
    queryables: dict
    typename: dict

    """

    for item in queryables:
        try:
            matching_typename = [key for key, value in typename.items() if
                                 value == item][0]
            queryable_value = queryables[matching_typename]
            queryables[item] = {
                "xpath": queryable_value["xpath"],
                "dbcol": queryable_value["dbcol"],
            }
        except IndexError:
            pass


def getqattr(obj, name):
    """Get value of an object, safely"""
    result = None
    try:
        item = getattr(obj, name)
        value = item()
        if "link" in name:  # create link format
            links = []
            for link in value:
                links.append(','.join(list(link)))
            result = '^'.join(links)
        else:
            result = value
    except TypeError:  # item is not callable
        try:
            result = datetime2iso8601(item)
        except AttributeError:  # item is not date(time)
            result = item
    except AttributeError:  # obj does not have a name property
        pass
    return result


def http_request(method, url, request=None, timeout=30):
    """Perform HTTP request"""
    if method == 'POST':
        return http_post(url, request, timeout=timeout)
    else:  # GET
        request = Request(url)
        request.add_header('User-Agent', 'pycsw (https://pycsw.org/)')
        return urlopen(request, timeout=timeout).read()


def bind_url(url):
    """binds an HTTP GET query string endpoint"""
    parsed_url = urlparse(url)
    if parsed_url.query == "":
        binder = "?"
    elif parsed_url.query.endswith("&"):
        binder = ""
    else:
        binder = "&"
    return "".join((parsed_url.geturl(), binder))


def ip_in_network_cidr(ip, net):
    """decipher whether IP is within CIDR range"""
    ipaddr = int(
        ''.join(['%02x' % int(x) for x in ip.split('.')]),
        16
    )
    netstr, bits = net.split('/')
    netaddr = int(
        ''.join(['%02x' % int(x) for x in netstr.split('.')]),
        16
    )
    mask = (0xffffffff << (32 - int(bits))) & 0xffffffff
    return (ipaddr & mask) == (netaddr & mask)


def ipaddress_in_whitelist(ipaddress, whitelist):
    """decipher whether IP is in IP whitelist

    IP whitelist is a list supporting:
    - single IP address (e.g. 192.168.0.1)
    - IP range using CIDR (e.g. 192.168.0/22)
    - IP range using subnet wildcard (e.g. 192.168.0.*, 192.168.*)

    """

    if ipaddress in whitelist:
        return True
    else:
        for white in whitelist:
            if white.find('/') != -1:  # CIDR
                if ip_in_network_cidr(ipaddress, white):
                    return True
            elif white.find('*') != -1:  # subnet wildcard
                    if ipaddress.startswith(white.split('*')[0]):
                        return True
    return False


def get_anytext(bag):
    """
    generate bag of text for free text searches
    accepts list of words, string of XML, or etree.Element
    """

    if isinstance(bag, list):  # list of words
        return ' '.join([_f for _f in bag if _f]).strip()
    else:  # xml
        if isinstance(bag, bytes) or isinstance(bag, str):
            # serialize to lxml
            bag = etree.fromstring(bag, PARSER)
        # get all XML element content
        return ' '.join([value.strip() for value in bag.xpath('//text()')])


# https://github.com/pallets/werkzeug/blob/778f482d1ac0c9e8e98f774d2595e9074e6984d7/werkzeug/utils.py#L253
def secure_filename(filename):
    r"""Pass it a filename and it will return a secure version of it.  This
    filename can then safely be stored on a regular file system and passed
    to :func:`os.path.join`.  The filename returned is an ASCII only string
    for maximum portability.

    On windows systems the function also makes sure that the file is not
    named after one of the special device files.

    >>> secure_filename("My cool movie.mov")
    'My_cool_movie.mov'
    >>> secure_filename("../../../etc/passwd")
    'etc_passwd'
    >>> secure_filename(u'i contain cool \xfcml\xe4uts.txt')
    'i_contain_cool_umlauts.txt'

    The function might return an empty filename.  It's your responsibility
    to ensure that the filename is unique and that you abort or
    generate a random filename if the function returned an empty one.

    .. versionadded:: 0.5

    :param filename: the filename to secure
    """
    if isinstance(filename, str):
        from unicodedata import normalize
        filename = normalize('NFKD', filename).encode('ascii', 'ignore')
        filename = filename.decode('ascii')
    for sep in os.path.sep, os.path.altsep:
        if sep:
            filename = filename.replace(sep, ' ')
    filename = str(_filename_ascii_strip_re.sub('', '_'.join(
                   filename.split()))).strip('._')

    # on nt a couple of special files are present in each folder.  We
    # have to ensure that the target file is not such a filename.  In
    # this case we prepend an underline
    if os.name == 'nt' and filename and \
       filename.split('.')[0].upper() in _windows_device_files:
        filename = '_' + filename

    return filename



# =================================================================
#
# Authors: Ricardo Garcia Silva <ricardo.garcia.silva@gmail.com>
#
# Copyright (c) 2016 Ricardo Garcia Silva
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================
"""pytest configuration file"""

import pytest


def pytest_configure(config):
    """Configure pytest.

    This function adds additional markers to pytest.

    """

    config.addinivalue_line(
        "markers",
        "functional: Run only functional tests"
    )
    config.addinivalue_line(
        "markers",
        "unit: Run only unit tests"
    )


def pytest_addoption(parser):
    """Add additional command-line parameters to pytest."""
    parser.addoption(
        "--database-backend",
        choices=["sqlite", "postgresql"],
        default="sqlite",
        help="Database backend to use when performing functional tests"
    )
    parser.addoption(
        "--database-user-postgresql",
        default="postgres",
        help="Username to use for creating and accessing local postgres "
             "databases used for functional tests."
    )
    parser.addoption(
        "--database-password-postgresql",
        default="",
        help="Password to use for creating and accessing local postgres "
             "databases used for functional tests."
    )
    parser.addoption(
        "--database-name-postgresql",
        default="test_pycsw",
        help="Name of the postgres database that is to be used for testing."
    )
    parser.addoption(
        "--database-host-postgresql",
        default="127.0.0.1",
        help="hostname or ip of the host that is running the postgres "
             "database server to use in testing."
    )
    parser.addoption(
        "--database-port-postgresql",
        default="5432",
        help="Port where the postgres server is listening for connections."
    )
    parser.addoption(
        "--pycsw-loglevel",
        default="warning",
        help="Log level for the pycsw server."
    )
    parser.addoption(
        "--functional-prefer-diffs",
        action="store_true",
        help="When running functional tests, compare results with their "
             "expected values by using diffs instead of XML canonicalisation "
             "(the default)."
    )
    parser.addoption(
        "--functional-save-results-directory",
        help="When running functional tests, save each test's result under "
             "the input directory path."
    )


@pytest.fixture(scope="session")
def log_level(request):
    """Log level to use when instantiating a new pycsw server.

    The value for this fixture is retrieved from the `--pycsw.loglevel`
    command-line parameter.

    """

    return request.config.getoption("pycsw_loglevel").upper()



# -*- coding: utf-8 -*-
# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#          Angelos Tzotsos <tzotsos@gmail.com>
#          Ricardo Garcia Silva <ricardo.garcia.silva@gmail.com>
#
# Copyright (c) 2019  Tom Kralidis
# Copyright (c) 2015 Angelos Tzotsos
# Copyright (c) 2017 Ricardo Garcia Silva
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import inspect
import logging
import os

from shapely.wkt import loads
try:
    from shapely.errors import ReadingError
except:
    from shapely.geos import ReadingError

from sqlalchemy import create_engine, func, __version__, select
from sqlalchemy.sql import text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import create_session

from pycsw.core import util
from pycsw.core.etree import etree
from pycsw.core.etree import PARSER

LOGGER = logging.getLogger(__name__)


class Repository(object):
    _engines = {}

    @classmethod
    def create_engine(clazz, url):
        '''
        SQL Alchemy engines are thread-safe and simple wrappers for connection pools

        https://groups.google.com/forum/#!topic/sqlalchemy/t8i3RSKZGb0

        To reduce startup time we can cache the engine as a class variable in the
        repository object and do database initialization once

        Engines are memoized by url
        '''
        if url not in clazz._engines:
            LOGGER.info('creating new engine: %s', url)
            engine = create_engine('%s' % url, echo=False)

            # load SQLite query bindings
            # This can be directly bound via events
            # for sqlite < 0.7, we need to to this on a per-connection basis
            if engine.name in ['sqlite', 'sqlite3'] and __version__ >= '0.7':
                from sqlalchemy import event
                @event.listens_for(engine, "connect")
                def connect(dbapi_connection, connection_rec):
                    create_custom_sql_functions(dbapi_connection)

            clazz._engines[url] = engine

        return clazz._engines[url]

    ''' Class to interact with underlying repository '''
    def __init__(self, database, context, app_root=None, table='records', repo_filter=None):
        ''' Initialize repository '''

        self.context = context
        self.filter = repo_filter
        self.fts = False

        # Don't use relative paths, this is hack to get around
        # most wsgi restriction...
        if (app_root and database.startswith('sqlite:///') and
            not database.startswith('sqlite:////')):
            database = database.replace('sqlite:///',
                       'sqlite:///%s%s' % (app_root, os.sep))

        self.engine = Repository.create_engine('%s' % database)

        base = declarative_base(bind=self.engine)

        LOGGER.info('binding ORM to existing database')

        self.postgis_geometry_column = None

        schema_name, table_name = table.rpartition(".")[::2]

        self.dataset = type(
            'dataset',
            (base,),
            {
                "__tablename__": table_name,
                "__table_args__": {
                    "autoload": True,
                    "schema": schema_name or None,
                },
            }
        )

        self.dbtype = self.engine.name

        self.session = create_session(self.engine)

        temp_dbtype = None

        if self.dbtype == 'postgresql':
            # check if PostgreSQL is enabled with PostGIS 1.x
            try:
                self.session.execute(select([func.postgis_version()]))
                temp_dbtype = 'postgresql+postgis+wkt'
                LOGGER.debug('PostgreSQL+PostGIS1+WKT detected')
            except Exception as err:
                LOGGER.exception('PostgreSQL+PostGIS1+WKT detection failed')

            # check if PostgreSQL is enabled with PostGIS 2.x
            try:
                self.session.execute('select(postgis_version())')
                temp_dbtype = 'postgresql+postgis+wkt'
                LOGGER.debug('PostgreSQL+PostGIS2+WKT detected')
            except Exception as err:
                LOGGER.exception('PostgreSQL+PostGIS2+WKT detection failed')

            # check if a native PostGIS geometry column exists
            try:
                result = self.session.execute(
                    "select f_geometry_column "
                    "from geometry_columns "
                    "where f_table_name = '%s' "
                    "and f_geometry_column != 'wkt_geometry' "
                    "limit 1;" % table_name
                )
                row = result.fetchone()
                self.postgis_geometry_column = str(row['f_geometry_column'])
                temp_dbtype = 'postgresql+postgis+native'
                LOGGER.debug('PostgreSQL+PostGIS+Native detected')
            except Exception as err:
                LOGGER.exception('PostgreSQL+PostGIS+Native not picked up: %s')

            # check if a native PostgreSQL FTS GIN index exists
            result = self.session.execute("select relname from pg_class where relname='fts_gin_idx'").scalar()
            self.fts = bool(result)
            LOGGER.debug('PostgreSQL FTS enabled: %r', self.fts)

        if temp_dbtype is not None:
            LOGGER.debug('%s support detected', temp_dbtype)
            self.dbtype = temp_dbtype

        if self.dbtype in ['sqlite', 'sqlite3']:  # load SQLite query bindings
            # <= 0.6 behaviour
            if not __version__ >= '0.7':
                self.connection = self.engine.raw_connection()
                create_custom_sql_functions(self.connection)

        LOGGER.info('setting repository queryables')
        # generate core queryables db and obj bindings
        self.queryables = {}

        for tname in self.context.model['typenames']:
            for qname in self.context.model['typenames'][tname]['queryables']:
                self.queryables[qname] = {}

                for qkey, qvalue in \
                self.context.model['typenames'][tname]['queryables'][qname].items():
                    self.queryables[qname][qkey] = qvalue

        # flatten all queryables
        # TODO smarter way of doing this
        self.queryables['_all'] = {}
        for qbl in self.queryables:
            self.queryables['_all'].update(self.queryables[qbl])

        self.queryables['_all'].update(self.context.md_core_model['mappings'])

    def _create_values(self, values):
        value_dict = {}
        for num, value in enumerate(values):
            value_dict['pvalue%d' % num] = value
        return value_dict

    def query_ids(self, ids):
        ''' Query by list of identifiers '''

        column = getattr(self.dataset, \
        self.context.md_core_model['mappings']['pycsw:Identifier'])

        query = self.session.query(self.dataset).filter(column.in_(ids))
        return self._get_repo_filter(query).all()

    def query_domain(self, domain, typenames, domainquerytype='list',
        count=False):
        ''' Query by property domain values '''

        domain_value = getattr(self.dataset, domain)

        if domainquerytype == 'range':
            LOGGER.info('Generating property name range values')
            query = self.session.query(func.min(domain_value),
                                       func.max(domain_value))
        else:
            if count:
                LOGGER.info('Generating property name frequency counts')
                query = self.session.query(getattr(self.dataset, domain),
                    func.count(domain_value)).group_by(domain_value)
            else:
                query = self.session.query(domain_value).distinct()
        return self._get_repo_filter(query).all()

    def query_insert(self, direction='max'):
        ''' Query to get latest (default) or earliest update to repository '''
        column = getattr(self.dataset, \
        self.context.md_core_model['mappings']['pycsw:InsertDate'])

        if direction == 'min':
            return self._get_repo_filter(self.session.query(func.min(column))).first()[0]
        # else default max
        return self._get_repo_filter(self.session.query(func.max(column))).first()[0]

    def query_source(self, source):
        ''' Query by source '''
        column = getattr(self.dataset, \
        self.context.md_core_model['mappings']['pycsw:Source'])

        query = self.session.query(self.dataset).filter(column == source)
        return self._get_repo_filter(query).all()

    def query(self, constraint, sortby=None, typenames=None,
        maxrecords=10, startposition=0):
        ''' Query records from underlying repository '''

        # run the raw query and get total
        if 'where' in constraint:  # GetRecords with constraint
            LOGGER.debug('constraint detected')
            query = self.session.query(self.dataset).filter(
            text(constraint['where'])).params(self._create_values(constraint['values']))
        else:  # GetRecords sans constraint
            LOGGER.debug('No constraint detected')
            query = self.session.query(self.dataset)

        total = self._get_repo_filter(query).count()

        if util.ranking_pass:  #apply spatial ranking
            #TODO: Check here for dbtype so to extract wkt from postgis native to wkt
            LOGGER.debug('spatial ranking detected')
            LOGGER.debug('Target WKT: %s', getattr(self.dataset, self.context.md_core_model['mappings']['pycsw:BoundingBox']))
            LOGGER.debug('Query WKT: %s', util.ranking_query_geometry)
            query = query.order_by(func.get_spatial_overlay_rank(getattr(self.dataset, self.context.md_core_model['mappings']['pycsw:BoundingBox']), util.ranking_query_geometry).desc())
            #trying to make this wsgi safe
            util.ranking_pass = False
            util.ranking_query_geometry = ''

        if sortby is not None:  # apply sorting
            LOGGER.debug('sorting detected')
            #TODO: Check here for dbtype so to extract wkt from postgis native to wkt
            sortby_column = getattr(self.dataset, sortby['propertyname'])

            if sortby['order'] == 'DESC':  # descending sort
                if 'spatial' in sortby and sortby['spatial']:  # spatial sort
                    query = query.order_by(func.get_geometry_area(sortby_column).desc())
                else:  # aspatial sort
                    query = query.order_by(sortby_column.desc())
            else:  # ascending sort
                if 'spatial' in sortby and sortby['spatial']:  # spatial sort
                    query = query.order_by(func.get_geometry_area(sortby_column))
                else:  # aspatial sort
                    query = query.order_by(sortby_column)

        # always apply limit and offset
        return [str(total), self._get_repo_filter(query).limit(
        maxrecords).offset(startposition).all()]

    def insert(self, record, source, insert_date):
        ''' Insert a record into the repository '''

        if isinstance(record.xml, bytes):
            LOGGER.debug('Decoding bytes to unicode')
            record.xml = record.xml.decode()

        try:
            self.session.begin()
            self.session.add(record)
            self.session.commit()
        except Exception as err:
            self.session.rollback()
            raise

    def update(self, record=None, recprops=None, constraint=None):
        ''' Update a record in the repository based on identifier '''

        if record is not None:
            identifier = getattr(record,
            self.context.md_core_model['mappings']['pycsw:Identifier'])
            xml = getattr(self.dataset,
            self.context.md_core_model['mappings']['pycsw:XML'])
            anytext = getattr(self.dataset,
            self.context.md_core_model['mappings']['pycsw:AnyText'])

        if recprops is None and constraint is None:  # full update
            LOGGER.debug('full update')
            update_dict = dict([(getattr(self.dataset, key),
            getattr(record, key)) \
            for key in record.__dict__.keys() if key != '_sa_instance_state'])

            try:
                self.session.begin()
                self._get_repo_filter(self.session.query(self.dataset)).filter_by(
                identifier=identifier).update(update_dict, synchronize_session='fetch')
                self.session.commit()
            except Exception as err:
                self.session.rollback()
                msg = 'Cannot commit to repository'
                LOGGER.exception(msg)
                raise RuntimeError(msg)
        else:  # update based on record properties
            LOGGER.debug('property based update')
            try:
                rows = rows2 = 0
                self.session.begin()
                for rpu in recprops:
                    # update queryable column and XML document via XPath
                    if 'xpath' not in rpu['rp']:
                        self.session.rollback()
                        raise RuntimeError('XPath not found for property %s' % rpu['rp']['name'])
                    if 'dbcol' not in rpu['rp']:
                        self.session.rollback()
                        raise RuntimeError('property not found for XPath %s' % rpu['rp']['name'])
                    rows += self._get_repo_filter(self.session.query(self.dataset)).filter(
                        text(constraint['where'])).params(self._create_values(constraint['values'])).update({
                            getattr(self.dataset,
                            rpu['rp']['dbcol']): rpu['value'],
                            'xml': func.update_xpath(str(self.context.namespaces),
                                   getattr(self.dataset,
                                   self.context.md_core_model['mappings']['pycsw:XML']),
                                   str(rpu)),
                        }, synchronize_session='fetch')
                    # then update anytext tokens
                    rows2 += self._get_repo_filter(self.session.query(self.dataset)).filter(
                        text(constraint['where'])).params(self._create_values(constraint['values'])).update({
                            'anytext': func.get_anytext(getattr(
                            self.dataset, self.context.md_core_model['mappings']['pycsw:XML']))
                        }, synchronize_session='fetch')
                self.session.commit()
                return rows
            except Exception as err:
                self.session.rollback()
                msg = 'Cannot commit to repository'
                LOGGER.exception(msg)
                raise RuntimeError(msg)

    def delete(self, constraint):
        ''' Delete a record from the repository '''

        try:
            self.session.begin()
            rows = self._get_repo_filter(self.session.query(self.dataset)).filter(
            text(constraint['where'])).params(self._create_values(constraint['values']))

            parentids = []
            for row in rows:  # get ids
                parentids.append(getattr(row,
                self.context.md_core_model['mappings']['pycsw:Identifier']))

            rows=rows.delete(synchronize_session='fetch')

            if rows > 0:
                LOGGER.debug('Deleting all child records')
                # delete any child records which had this record as a parent
                rows += self._get_repo_filter(self.session.query(self.dataset)).filter(
                    getattr(self.dataset,
                    self.context.md_core_model['mappings']['pycsw:ParentIdentifier']).in_(parentids)).delete(
                    synchronize_session='fetch')

            self.session.commit()
        except Exception as err:
            self.session.rollback()
            msg = 'Cannot commit to repository'
            LOGGER.exception(msg)
            raise RuntimeError(msg)

        return rows

    def _get_repo_filter(self, query):
        ''' Apply repository wide side filter / mask query '''
        if self.filter is not None:
            return query.filter(text(self.filter))
        return query


def create_custom_sql_functions(connection):
    """Register custom functions on the database connection."""

    inspect_function = inspect.getfullargspec

    for function_object in [
        query_spatial,
        update_xpath,
        util.get_anytext,
        get_geometry_area,
        get_spatial_overlay_rank
    ]:
        argspec = inspect_function(function_object)
        connection.create_function(
            function_object.__name__,
            len(argspec.args),
            function_object
        )


def query_spatial(bbox_data_wkt, bbox_input_wkt, predicate, distance):
    """Perform spatial query

    Parameters
    ----------
    bbox_data_wkt: str
        Well-Known Text representation of the data being queried
    bbox_input_wkt: str
        Well-Known Text representation of the input being queried
    predicate: str
        Spatial predicate to use in query
    distance: int or float or str
        Distance parameter for when using either of ``beyond`` or ``dwithin``
        predicates.

    Returns
    -------
    str
        Either ``true`` or ``false`` depending on the result of the spatial
        query

    Raises
    ------
    RuntimeError
        If an invalid predicate is used

    """

    try:
        bbox1 = loads(bbox_data_wkt.split(';')[-1])
        bbox2 = loads(bbox_input_wkt)
        if predicate == 'bbox':
            result = bbox1.intersects(bbox2)
        elif predicate == 'beyond':
            result = bbox1.distance(bbox2) > float(distance)
        elif predicate == 'contains':
            result = bbox1.contains(bbox2)
        elif predicate == 'crosses':
            result = bbox1.crosses(bbox2)
        elif predicate == 'disjoint':
            result = bbox1.disjoint(bbox2)
        elif predicate == 'dwithin':
            result = bbox1.distance(bbox2) <= float(distance)
        elif predicate == 'equals':
            result = bbox1.equals(bbox2)
        elif predicate == 'intersects':
            result = bbox1.intersects(bbox2)
        elif predicate == 'overlaps':
            result = bbox1.intersects(bbox2) and not bbox1.touches(bbox2)
        elif predicate == 'touches':
            result = bbox1.touches(bbox2)
        elif predicate == 'within':
            result = bbox1.within(bbox2)
        else:
            raise RuntimeError(
                'Invalid spatial query predicate: %s' % predicate)
    except (AttributeError, ValueError, ReadingError, TypeError):
        result = False
    return "true" if result else "false"


def update_xpath(nsmap, xml, recprop):
    """Update XML document XPath values"""

    if isinstance(xml, bytes) or isinstance(xml, str):
        # serialize to lxml
        xml = etree.fromstring(xml, PARSER)

    recprop = eval(recprop)
    nsmap = eval(nsmap)
    try:
        nodes = xml.xpath(recprop['rp']['xpath'], namespaces=nsmap)
        if len(nodes) > 0:  # matches
            for node1 in nodes:
                if node1.text != recprop['value']:  # values differ, update
                    node1.text = recprop['value']
    except Exception as err:
        LOGGER.warning(err)
        raise RuntimeError('ERROR: %s' % str(err))

    return etree.tostring(xml)


def get_geometry_area(geometry):
    """Derive area of a given geometry"""
    try:
        if geometry is not None:
            return str(loads(geometry).area)
        return '0'
    except:
        return '0'


def get_spatial_overlay_rank(target_geometry, query_geometry):
    """Derive spatial overlay rank for geospatial search as per Lanfear (2006)
    http://pubs.usgs.gov/of/2006/1279/2006-1279.pdf"""

    #TODO: Add those parameters to config file
    kt = 1.0
    kq = 1.0
    if target_geometry is not None and query_geometry is not None:
        try:
            q_geom = loads(query_geometry)
            t_geom = loads(target_geometry)
            Q = q_geom.area
            T = t_geom.area
            if any(item == 0.0 for item in [Q, T]):
                LOGGER.warning('Geometry has no area')
                return '0'
            X = t_geom.intersection(q_geom).area
            if kt == 1.0 and kq == 1.0:
                LOGGER.debug('Spatial Rank: %s', str((X/Q)*(X/T)))
                return str((X/Q)*(X/T))
            else:
                LOGGER.debug('Spatial Rank: %s', str(((X/Q)**kq)*((X/T)**kt)))
                return str(((X/Q)**kq)*((X/T)**kt))
        except Exception as err:
            LOGGER.warning('Cannot derive spatial overlay ranking %s', err)
            return '0'
    return '0'

